@@ -1,4 +1,4 @@
-/*  gss-utils.h,v 1.11 2005/05/04 23:04:48 eagle Exp
+/*  $Id: gss-utils.h 2313 2005-12-22 22:16:43Z rra $
 **
 **  Utility functions for remctl.
 **
@@ -12,13 +12,13 @@
 #ifndef GSS_UTILS_H
 #define GSS_UTILS_H 1
 
-#include <gssapi/gssapi_generic.h>
+#include "config.h"
 
-/* Replacement functions that check the results of memory allocation and never
-   fail from the perspective of the rest of the program. */
-void *smalloc(int);
-void *srealloc(void* ptr, size_t size);
-char *sstrdup(const char* s1);
+#ifdef HAVE_GSSAPI_H
+# include <gssapi.h>
+#else
+# include <gssapi/gssapi_generic.h>
+#endif
 
 /* Lowercase a string. */
 void lowercase(char *string);
