@@ -1,4 +1,4 @@
-/*  remctl.c,v 1.17 2005/05/04 23:04:48 eagle Exp
+/*  $Id: remctl.c 2315 2005-12-22 22:36:17Z rra $
 **
 **  The client for a "K5 sysctl" - a service for remote execution of 
 **  predefined commands.  Access is authenticated via GSSAPI Kerberos 5, 
@@ -22,11 +22,37 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#include <gssapi/gssapi_generic.h>
-#include "gss-utils.h"
+#ifdef HAVE_GSSAPI_H
+# include <gssapi.h>
+#else
+# include <gssapi/gssapi_generic.h>
+#endif
 
-int verbose;       /* Turns on debugging output. */
-int use_syslog;    /* Toggle for sysctl vs stdout/stderr. */
+#include "gss-utils.h"
+#include "messages.h"
+#include "xmalloc.h"
+
+/* Handle compatibility to older versions of MIT Kerberos. */
+#ifndef HAVE_GSS_RFC_OIDS
+# define GSS_C_NT_USER_NAME gss_nt_user_name
+#endif
+
+/* Heimdal provides a nice #define for this. */
+#if !HAVE_DECL_GSS_KRB5_MECHANISM
+# include <gssapi/gssapi_krb5.h>
+# define GSS_KRB5_MECHANISM gss_mech_krb5
+#endif
+
+/* Usage message. */
+static const char usage_message[] = "\
+Usage: remctl <options> <host> <type> <service> <parameters>\n\
+\n\
+Options:\n\
+    -d            Debugging level of output\n\
+    -h            Display this help\n\
+    -p <port>     remctld port (default: 4444)\n\
+    -s <service>  remctld service principal (default: host/<host>)\n\
+    -v            Display the version of remctl\n";
 
 /* These are for storing either the socket for communication with client
    or the streams that talk to the network in case of inetd/tcpserver. */
@@ -42,57 +68,17 @@ gss_buffer_t empty_token = &empty_token_buf;
 **  Display the usage message for remctl.
 */
 static void
-usage(void)
+usage(int status)
 {
-    static const char usage[] = "\
-Usage: remctl <options> host type service <parameters>\n\
-Options:\n\
-\t-s     remctld service principal (default host/machine.stanford.edu)\n\
-\t-v     debugging level of output\n";
-    fprintf(stderr, usage);
-    exit(1);
-}
-
-
-/*
-**  Parse an OID string and determine the mechanism type, returning it in the
-**  oid argument.  Prints an error message if there are any problems.
-*/
-static void
-parse_oid(const char *mechanism, gss_OID *oid)
-{
-    char *mechstr = 0, *cp;
-    gss_buffer_desc tok;
-    OM_uint32 maj_stat, min_stat;
-
-    if (isdigit(mechanism[0])) {
-        mechstr = smalloc(strlen(mechanism) + 5);
-        if (!mechstr) {
-            fprintf(stderr, "Couldn't allocate mechanism scratch!\n");
-            return;
-        }
-        sprintf(mechstr, "{ %s }", mechanism);
-        for (cp = mechstr; *cp; cp++)
-            if (*cp == '.')
-                *cp = ' ';
-        tok.value = mechstr;
-    } else
-        tok.value = (char *) mechanism;
-    tok.length = strlen(tok.value);
-    maj_stat = gss_str_to_oid(&min_stat, &tok, oid);
-    if (maj_stat != GSS_S_COMPLETE) {
-        display_status("while str_to_oid", maj_stat, min_stat);
-        return;
-    }
-    if (mechstr)
-        free(mechstr);
+    fprintf((status == 0) ? stdout : stderr, "%s", usage_message);
+    exit(status);
 }
 
 
 /*
 **  Opens a TCP connection to the given hostname and port.  The host name is
 **  resolved by gethostbyname and the socket is opened and connected.  Returns
-**  the file descriptor or -1 on failure, printing an error message.
+**  the file descriptor or dies on failure.
 */
 static int
 connect_to_server(char *host, unsigned short port)
@@ -101,24 +87,20 @@ connect_to_server(char *host, unsigned short port)
     struct hostent *hp;
     int s;
 
-    if ((hp = gethostbyname(host)) == NULL) {
-        fprintf(stderr, "Unknown host: %s\n", host);
-        return -1;
-    }
+    hp = gethostbyname(host);
+    if (hp == NULL)
+        die("unknown host: %s", host);
 
     saddr.sin_family = hp->h_addrtype;
     memcpy((char *)&saddr.sin_addr, hp->h_addr, sizeof(saddr.sin_addr));
     saddr.sin_port = htons(port);
 
-    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
-        perror("creating socket");
-        return -1;
-    }
-    if (connect(s, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
-        perror("connecting to server");
-        (void)close(s);
-        return -1;
-    }
+    s = socket(AF_INET, SOCK_STREAM, 0);
+    if (s < 0)
+        sysdie("error creating socket");
+    if (connect(s, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
+        sysdie("cannot connect to %s (port %hu)", host, port);
+
     return s;
 }
 
@@ -129,7 +111,6 @@ connect_to_server(char *host, unsigned short port)
 **  arguments.  Returns 0 on success, -1 on failure and prints an error
 **  message.
 **
-**
 **  The service name is imported as a GSSAPI name and a GSS-API context is
 **  established with the corresponding service.  The service should be
 **  listening on the TCP connection s.  The default GSS-API mechanism is used,
@@ -139,27 +120,24 @@ static int
 client_establish_context(char *service_name, gss_ctx_id_t *gss_context,
                          OM_uint32 *ret_flags)
 {
-    gss_OID oid = GSS_C_NULL_OID;
     gss_buffer_desc send_tok, recv_tok, *token_ptr;
     gss_name_t target_name;
     OM_uint32 maj_stat, min_stat, init_sec_min_stat;
     int token_flags;
 
-    parse_oid("1.2.840.113554.1.2.2", &oid);
-
     /* Import the name into target_name.  Use send_tok to save local variable
        space. */
     send_tok.value = service_name;
     send_tok.length = strlen(service_name) + 1;
-    maj_stat = gss_import_name(&min_stat, &send_tok,
-                               (gss_OID) gss_nt_user_name, &target_name);
+    maj_stat = gss_import_name(&min_stat, &send_tok, GSS_C_NT_USER_NAME,
+                               &target_name);
     if (maj_stat != GSS_S_COMPLETE) {
         display_status("while parsing name", maj_stat, min_stat);
         return -1;
     }
 
     if (send_token(TOKEN_NOOP | TOKEN_CONTEXT_NEXT, empty_token) < 0) {
-        fprintf(stderr, "Can't send initial token to server\n");
+        warn("cannot send initial token to server");
         gss_release_name(&min_stat, &target_name);
         return -1;
     }
@@ -184,7 +162,7 @@ client_establish_context(char *service_name, gss_ctx_id_t *gss_context,
                                         GSS_C_NO_CREDENTIAL, 
                                         gss_context, 
                                         target_name, 
-                                        oid, 
+                                        (const gss_OID) GSS_KRB5_MECHANISM,
                                         GSS_C_MUTUAL_FLAG | GSS_C_REPLAY_FLAG, 
                                         0, 
                                         NULL,     /* no channel bindings */
@@ -198,9 +176,7 @@ client_establish_context(char *service_name, gss_ctx_id_t *gss_context,
             gss_release_buffer(&min_stat, &recv_tok);
 
         if (send_tok.length != 0) {
-            if (verbose)
-                printf("Sending init_sec_context token (size=%d)...\n",
-                       send_tok.length);
+            debug("sending init_sec_context token, size=%d", send_tok.length);
             if (send_token(TOKEN_CONTEXT, &send_tok) < 0) {
                 gss_release_buffer(&min_stat, &send_tok);
                 gss_release_name(&min_stat, &target_name);
@@ -221,21 +197,17 @@ client_establish_context(char *service_name, gss_ctx_id_t *gss_context,
         }
 
         if (maj_stat == GSS_S_CONTINUE_NEEDED) {
-            if (verbose)
-                printf("continue needed...");
+            debug("continue needed");
             if (recv_token(&token_flags, &recv_tok) < 0) {
-                fprintf(stderr, "No token received from server\n");
+                warn("no token received from server");
                 gss_release_name(&min_stat, &target_name);
                 return -1;
             }
             token_ptr = &recv_tok;
         }
-        if (verbose)
-            printf("\n");
     } while (maj_stat == GSS_S_CONTINUE_NEEDED);
 
     gss_release_name(&min_stat, &target_name);
-    gss_release_oid(&min_stat, &oid);
 
     return 0;
 }
@@ -264,7 +236,7 @@ process_request(gss_ctx_id_t context, OM_uint32 argc, char **argv)
         msglength += strlen(*cp++);
     }
     msglength += (argc + 1) * sizeof(OM_uint32);
-    msg = smalloc(msglength);
+    msg = xmalloc(msglength);
 
     /* Iterators over buffers */
     mp = msg;
@@ -288,10 +260,12 @@ process_request(gss_ctx_id_t context, OM_uint32 argc, char **argv)
         cp++;
     }
 
-    if (gss_sendmsg(context, TOKEN_DATA, msg, msglength) < 0)
-        return (-1);
-
-    return (0);
+    if (gss_sendmsg(context, TOKEN_DATA, msg, msglength) < 0) {
+        free(msg);
+        return -1;
+    }
+    free(msg);
+    return 0;
 }
 
 
@@ -313,7 +287,7 @@ process_response(gss_ctx_id_t context, OM_uint32* errorcode)
     int token_flags;
 
     if (gss_recvmsg(context, &token_flags, &msg, &msglength) < 0)
-        return (-1);
+        return -1;
 
     cp = msg;
 
@@ -321,20 +295,17 @@ process_response(gss_ctx_id_t context, OM_uint32* errorcode)
     memcpy(&network_ordered, cp, sizeof(OM_uint32));
     *errorcode = ntohl(network_ordered);
     cp += sizeof(OM_uint32);
-    if (verbose)
-        printf("Return code: %d\n", *errorcode);
+    debug("return code: %d", *errorcode);
 
     /* Get the message length */
     memcpy(&network_ordered, cp, sizeof(OM_uint32));
     cp += sizeof(OM_uint32);
     bodylength = ntohl(network_ordered);
-    if (bodylength > MAXBUFFER || bodylength > msglength) {
-        fprintf(stderr, "Data unpacking error while processing response\n");
-        exit(-1);
-    }
+    if (bodylength > MAXBUFFER || bodylength > msglength)
+        die("data unpacking error while processing response");
 
     /* Get the message body */
-    body = smalloc(bodylength + 1);
+    body = xmalloc(bodylength + 1);
     memcpy(body, cp, bodylength);
     body[bodylength] = '\0';
     
@@ -351,100 +322,95 @@ process_response(gss_ctx_id_t context, OM_uint32* errorcode)
 **  call process_request and process_response.
 */
 int
-main(int argc, char ** argv)
+main(int argc, char *argv[])
 {
-    char service_name[500];
+    int option;
     char *server_host;
-    struct hostent* hostinfo;
+    struct hostent *hostinfo;
+    char *service_name = NULL;
     unsigned short port = 4444;
     OM_uint32 ret_flags, maj_stat, min_stat;
     int s;
     OM_uint32 errorcode = 0;
-
     gss_ctx_id_t context;
     gss_buffer_desc out_buf;
 
-    service_name[0] = '\0';
-    verbose = 0;
-    use_syslog = 0;
+    /* Set up logging and identity. */
+    message_program_name = "remctl";
 
-    /* Parse arguments. */
-    argc--;
-    argv++;
-    while (argc) {
-        if (strcmp(*argv, "-p") == 0) {
-            argc--;
-            argv++;
-            if (!argc)
-                usage();
-            port = atoi(*argv);
-        } else if (strcmp(*argv, "-s") == 0) {
-            argc--;
-            argv++;
-            if (!argc)
-                usage();
-            strncpy(service_name, *argv, sizeof(service_name));
-        } else if (strcmp(*argv, "-v") == 0) {
-            verbose = 1;
-        } else
+    /* Parse options.  The + tells GNU getopt to stop option parsing at the
+       first non-argument rather than proceeding on to find options anywhere.
+       Without this, it's hard to call remote programs that take options.
+       Non-GNU getopt will treat the + as a supported option, which is handled
+       below. */
+    while ((option = getopt(argc, argv, "+dhp:s:v")) != EOF) {
+        switch (option) {
+        case 'd':
+            message_handlers_debug(1, message_log_stderr);
             break;
-        argc--;
-        argv++;
+        case 'h':
+            usage(0);
+            break;
+        case 'p':
+            port = atoi(optarg);
+            break;
+        case 's':
+            service_name = optarg;
+            break;
+        case 'v':
+            printf("%s\n", PACKAGE_STRING);
+            exit(0);
+            break;
+        case '+':
+            fprintf(stderr, "%s: invalid option -- +\n", argv[0]);
+        default:
+            usage(1);
+            break;
+        }
     }
+    argc -= optind;
+    argv += optind;
     if (argc < 3)
-        usage();
-
+        usage(1);
     server_host = *argv++;
     argc--;
 
-    if (strlen(service_name) == 0) {
-        if ((hostinfo = gethostbyname(server_host)) == NULL) {
-            fprintf(stderr, "Can't resolve given hostname\n");
-            exit(1);
-        }
-
-        lowercase(hostinfo->h_name);
+    if (service_name == NULL) {
+        hostinfo = gethostbyname(server_host);
+        if (hostinfo == NULL)
+            die("cannot resolve host %s", server_host);
+        service_name = xmalloc(strlen("host/") + strlen(hostinfo->h_name) + 1);
         strcpy(service_name, "host/");
         strcat(service_name, hostinfo->h_name);
+        lowercase(service_name);
     }
 
     /* Open connection. */
-    if ((s = connect_to_server(server_host, port)) < 0)
-        return -1;
-
+    s = connect_to_server(server_host, port);
     READFD = s;
     WRITEFD = s;
 
     /* Establish context. */
-    if (client_establish_context(service_name, &context, &ret_flags) < 0) {
-        fprintf(stderr, "Can't establish GSS-API context\n");
-        close(s);
-        return -1;
-    }
+    if (client_establish_context(service_name, &context, &ret_flags) < 0)
+        die("cannot establish GSSAPI context");
 
     /* Display the flags. */
-    if (verbose)
-        display_ctx_flags(ret_flags);
+    display_ctx_flags(ret_flags);
 
-    if (process_request(context, argc, argv) < 0) {
-        fprintf(stderr, "Can't process request\n");
-        exit(1);
-    }
+    /* Do the work. */
+    if (process_request(context, argc, argv) < 0)
+        die("cannot process request");
+    if (process_response(context, &errorcode) < 0)
+        die("cannot process response");
 
-    if (process_response(context, &errorcode) < 0) {
-        fprintf(stderr, "Can't process response\n");
-        exit(1);
-    }
-
-    /* Delete context. */
+    /* Delete context and shut down cleanly. */
     maj_stat = gss_delete_sec_context(&min_stat, &context, &out_buf);
     if (maj_stat != GSS_S_COMPLETE) {
         display_status("while deleting context", maj_stat, min_stat);
         close(s);
         gss_delete_sec_context(&min_stat, &context, GSS_C_NO_BUFFER);
-        return -1;
+        exit(-1);
     }
-
     gss_release_buffer(&min_stat, &out_buf);
     close(s);
 
