@@ -1,4 +1,4 @@
-/*  remctld.c,v 1.40 2005/05/11 04:45:37 eagle Exp
+/*  $Id: remctld.c 2327 2005-12-24 07:38:38Z rra $
 **
 **  The deamon for a "K5 sysctl" - a service for remote execution of
 **  predefined commands.  Access is authenticated via GSSAPI Kerberos 5,
@@ -11,6 +11,8 @@
 **  See README for copyright and licensing information.
 */
 
+#include "config.h"
+
 #include <ctype.h>
 #include <dirent.h>
 #include <errno.h>
@@ -29,12 +31,34 @@
 #include <time.h>
 #include <unistd.h>
 
-#include <gssapi/gssapi_generic.h>
-#include "gss-utils.h"
-#include "vector.h"
+#ifdef HAVE_GSSAPI_H
+# include <gssapi.h>
+#else
+# include <gssapi/gssapi_generic.h>
+#endif
 
-int verbose;       /* Turns on debugging output. */
-int use_syslog;    /* Toggle for sysctl vs stdout/stderr. */
+#include "gss-utils.h"
+#include "messages.h"
+#include "vector.h"
+#include "xmalloc.h"
+
+/* Handle compatibility to older versions of MIT Kerberos. */
+#ifndef HAVE_GSS_RFC_OIDS
+# define GSS_C_NT_USER_NAME gss_nt_user_name
+#endif
+
+/* Usage message. */
+static const char usage_message[] = "\
+Usage: remctld <options>\n\
+\n\
+Options:\n\
+    -d            Log debugging information to syslog\n\
+    -f <file>     Config file (default: " CONFIG_FILE ")\n\
+    -h            Display this help\n\
+    -m            Stand-alone daemon mode, meant mostly for testing\n\
+    -s <service>  Service principal to use (default: host/<host>)\n\
+    -v            Display the version of remctld\n\
+    -p <port>     Port to use, only for standalone mode (default: 4444)\n";
 
 /* These are for storing either the socket for communication with client
    or the streams that talk to the network in case of inetd/tcpserver. */
@@ -63,20 +87,13 @@ struct config {
 **  Display the usage message for remctld.
 */
 static void
-usage(void)
+usage(int status)
 {
-    static const char usage[] = "\
-Usage: remctld <options>\n\
-Options:\n\
-\t-s service       K5 servicename to run as (default host/machine.stanford.edu)\
-\t-f conffile      the default is ./remctl.conf\n\
-\t-v               debugging level of output\n\
-\t-m               standalone single connection mode, meant for testing only\n\
-\t-p port          only for standalone mode. default 4444\n";
-
-    fprintf(stderr, usage);
-    syslog(LOG_ERR, "invalid usage");
-    exit(1);
+    fprintf((status == 0) ? stdout : stderr, usage_message);
+    if (status == 0)
+        exit(0);
+    else
+        die("invalid usage");
 }
 
 
@@ -96,19 +113,21 @@ create_socket(unsigned short port)
     saddr.sin_port = htons(port);
     saddr.sin_addr.s_addr = INADDR_ANY;
 
-    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
-        perror("creating socket");
+    s = socket(AF_INET, SOCK_STREAM, 0);
+    if (s < 0) {
+        syswarn("error creating socket");
         return -1;
     }
+
     /* Let the socket be reused right away */
     setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
     if (bind(s, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
-        perror("binding socket");
+        syswarn("error binding socket");
         close(s);
         return -1;
     }
     if (listen(s, 5) < 0) {
-        perror("listening on socket");
+        syswarn("error listening on socket");
         close(s);
         return -1;
     }
@@ -131,8 +150,8 @@ server_acquire_creds(char *service_name, gss_cred_id_t *server_creds)
     name_buf.value = service_name;
     name_buf.length = strlen(name_buf.value) + 1;
 
-    maj_stat = gss_import_name(&min_stat, &name_buf,
-                               (gss_OID) gss_nt_user_name, &server_name);
+    maj_stat = gss_import_name(&min_stat, &name_buf, GSS_C_NT_USER_NAME,
+                               &server_name);
 
     if (maj_stat != GSS_S_COMPLETE) {
         display_status("while importing name", maj_stat, min_stat);
@@ -172,7 +191,6 @@ server_establish_context(gss_cred_id_t server_creds, gss_ctx_id_t *context,
     gss_name_t client;
     gss_OID doid;
     OM_uint32 maj_stat, min_stat, acc_sec_min_stat;
-    gss_buffer_desc oid_name;
     int token_flags;
 
     if (recv_token(&token_flags, &recv_tok) < 0)
@@ -180,8 +198,7 @@ server_establish_context(gss_cred_id_t server_creds, gss_ctx_id_t *context,
 
     gss_release_buffer(&min_stat, &recv_tok);
     if (!(token_flags & TOKEN_NOOP)) {
-        syslog(LOG_ERR, "Expected NOOP token, got %d token instead\n",
-               token_flags);
+        warn("expected NOOP token, got %d token instead", token_flags);
         return -1;
     }
 
@@ -189,18 +206,13 @@ server_establish_context(gss_cred_id_t server_creds, gss_ctx_id_t *context,
 
     if (!(token_flags & TOKEN_CONTEXT_NEXT)) {
         client_name->length = *ret_flags = 0;
-        syslog(LOG_DEBUG,"Accepted unauthenticated connection.\n");
+        debug("accepted unauthenticated connection");
     } else {
         do {
             if (recv_token(&token_flags, &recv_tok) < 0)
                 return -1;
-
-            if (verbose) {
-                syslog(LOG_DEBUG, "Received token (size=%d): \n", 
-                       recv_tok.length);
-                print_token(&recv_tok);
-            }
-
+            debug("received token (size=%d)", recv_tok.length);
+            print_token(&recv_tok);
             maj_stat = gss_accept_sec_context(&acc_sec_min_stat, 
                                               context, 
                                               server_creds, 
@@ -217,17 +229,14 @@ server_establish_context(gss_cred_id_t server_creds, gss_ctx_id_t *context,
             gss_release_buffer(&min_stat, &recv_tok);
 
             if (send_tok.length != 0) {
-                if (verbose) {
-                    syslog(LOG_DEBUG,
-                           "Sending accept_sec_context token (size=%d):\n",
-                           send_tok.length);
-                    print_token(&send_tok);
-                }
+                debug("sending accept_sec_context token (size=%d)",
+                      send_tok.length);
+                print_token(&send_tok);
                 if (send_token(TOKEN_CONTEXT, &send_tok) < 0) {
-                    syslog(LOG_ERR, "failure sending token\n");
+                    warn("failure sending token");
+                    gss_release_buffer(&min_stat, &send_tok);
                     return -1;
                 }
-
                 gss_release_buffer(&min_stat, &send_tok);
             }
             if (maj_stat != GSS_S_COMPLETE
@@ -240,28 +249,11 @@ server_establish_context(gss_cred_id_t server_creds, gss_ctx_id_t *context,
                 return -1;
             }
 
-            if (verbose) {
-                if (maj_stat == GSS_S_CONTINUE_NEEDED)
-                    syslog(LOG_DEBUG,"continue needed...\n");
-                else
-                    syslog(LOG_DEBUG,"\n");
-            }
+            if (maj_stat == GSS_S_CONTINUE_NEEDED)
+                debug("continue needed");
         } while (maj_stat == GSS_S_CONTINUE_NEEDED);
 
-
-        if (verbose) {
-            display_ctx_flags(*ret_flags);
-
-            maj_stat = gss_oid_to_str(&min_stat, doid, &oid_name);
-            if (maj_stat != GSS_S_COMPLETE) {
-                display_status("while converting oid->string", maj_stat,
-                               min_stat);
-                return -1;
-            }
-            syslog(LOG_DEBUG,"Accepted connection using mechanism OID %.*s.\n",
-                   (int)oid_name.length, (char *)oid_name.value);
-            gss_release_buffer(&min_stat, &oid_name);
-        }
+        display_ctx_flags(*ret_flags);
 
         maj_stat = gss_display_name(&min_stat, client, client_name, &doid);
         if (maj_stat != GSS_S_COMPLETE) {
@@ -280,51 +272,70 @@ server_establish_context(gss_cred_id_t server_creds, gss_ctx_id_t *context,
 
 
 /*
-**  Reads a file into a character buffer, returning the newly allocated buffer
-**  or NULL on failure.
+**  Handles an include request for either read_conf_file or acl_check_file.
+**  Takes the vector that represents the include directive, the current file,
+**  the line number, the function to call for each included file, and a piece
+**  of data to pass to that function.  Handles including either files or
+**  directories.
+**
+**  If the function returns a value less than -1, return its return code.
+**  Otherwise, return the greatest of all status codes returned by the
+**  functions.  Also returns -2 for any error in processing the include
+**  directive.
 */
-static char *
-read_file(const char *file_name)
+static int
+handle_include(struct vector *line, const char *file, int lineno,
+               int (*function)(void *, const char *), void *data)
 {
-    int fd, count;
-    struct stat stat_buf;
-    int length;
-    char* buf;
+    const char *included = line->strings[1];
+    struct stat st;
 
-    if ((fd = open(file_name, O_RDONLY, 0)) < 0) {
-        syslog(LOG_ERR, "Can't open file %s: %m\n", file_name);
-        return NULL;
+    /* Validate the directive. */
+    if (line->count != 2 || strcmp(line->strings[0], "include") != 0) {
+        warn("%s:%d: parse error", file, lineno);
+        return -2;
     }
-    if (fstat(fd, &stat_buf) < 0) {
-        return NULL;
+    if (strcmp(included, file) == 0) {
+        warn("%s:%d: %s recursively included", file, lineno, file);
+        return -2;
     }
-    length = stat_buf.st_size;
-
-    if (length == 0) {
-        return NULL;
+    if (stat(included, &st) < 0) {
+        warn("%s:%d: included file %s not found", file, lineno, included);
+        return -2;
     }
 
-    if ((buf = smalloc(length + 1)) == 0) {
-        syslog(LOG_ERR, "Couldn't allocate %d byte buffer for reading file\n",
-               length);
-        return NULL;
-    }
+    /* If it's a directory, include everything in the directory that doesn't
+       contain a period.  Otherwise, just include the one file. */
+    if (!S_ISDIR(st.st_mode)) {
+        return (*function)(data, included);
+    } else {
+        DIR *dir;
+        struct dirent *entry;
+        int status = -1;
+        int last;
 
-    count = read(fd, buf, length);
-    if (count < 0) {
-        syslog(LOG_ERR, "Problem during read in read_file\n");
-        return NULL;
-    }
-    if (count < length) {
-        syslog(LOG_ERR, "Error: only read in %d bytes, expected %d\n",
-               count, length);
-        return NULL;
-    }
+        dir = opendir(included);
+        while ((entry = readdir(dir)) != NULL) {
+            char *path;
+            size_t length;
 
-    buf[length] = '\0';
-    close(fd);
-
-    return buf;
+            if (strchr(entry->d_name, '.') != NULL)
+                continue;
+            length = strlen(included) + 1 + strlen(entry->d_name) + 1;
+            path = xmalloc(length);
+            snprintf(path, length, "%s/%s", included, entry->d_name);
+            last = (*function)(data, path);
+            free(path);
+            if (last < -1) {
+                closedir(dir);
+                return status;
+            }
+            if (last > status)
+                status = last;
+        }
+        closedir(dir);
+        return status;
+    }
 }
 
 
@@ -336,51 +347,66 @@ read_file(const char *file_name)
 **  config is populated with the parsed configuration file.  Empty lines and
 **  lines beginning with # are ignored.  Each line is divided into fields,
 **  separated by spaces.  The fields are defined by struct confline.  Lines
-**  ending in backslash are continued on the next line.
+**  ending in backslash are continued on the next line.  config is passed in
+**  as a void * so that read_conf_file and acl_check_file can use common
+**  include handling code.
 **
 **  As a special case, include <file> will call read_conf_file recursively to
 **  parse an included file (or, if <file> is a directory, every file in that
 **  directory that doesn't contain a period).
 **
-**  Returns 0 on success and -1 on error, reporting an error message.
+**  Returns 0 on success and -2 on error, reporting an error message.
 */
 static int
-read_conf_file(struct config *config, const char *name)
+read_conf_file(void *data, const char *name)
 {
+    struct config *config = data;
     FILE *file;
     char *buffer, *p;
     size_t bufsize, length, size, count, i, arg_i;
-    struct vector *line;
-    struct confline *confline;
+    int s;
+    struct vector *line = NULL;
+    struct confline *confline = NULL;
     size_t lineno = 0;
+    DIR *dir = NULL;
 
     bufsize = 1024;
-    buffer = smalloc(bufsize);
+    buffer = xmalloc(bufsize);
     file = fopen(name, "r");
     if (file == NULL) {
         free(buffer);
-        syslog(LOG_ERR, "cannot open config file %s: %m", name);
-        return -1;
+        syswarn("cannot open config file %s", name);
+        return -2;
     }
     while (fgets(buffer, bufsize, file) != NULL) {
         length = strlen(buffer);
+        if (length == 2 && buffer[length - 1] != '\n') {
+            warn("%s:%d: no final newline", name, lineno);
+            goto fail;
+        }
+        if (length < 2)
+            continue;
 
         /* Allow for long lines and continuation lines.  As long as we've
            either filled the buffer or have a line ending in a backslash, we
            keep reading more data.  If we filled the buffer, increase it by
            another 1KB; otherwise, back up and write over the backslash and
            newline. */
-        while (length > 2 && (buffer[length - 1] != '\n'
-                              || buffer[length - 2] == '\\')) {
-            if (buffer[length - 1] != '\n') {
+        p = buffer + length - 2;
+        while (length > 2 && (p[1] != '\n' || p[0] == '\\')) {
+            if (p[1] != '\n') {
                 bufsize += 1024;
-                buffer = srealloc(buffer, bufsize);
+                buffer = xrealloc(buffer, bufsize);
             } else {
                 length -= 2;
+                lineno++;
+            }
+            if (fgets(buffer + length, bufsize - length, file) == NULL) {
+                warn("%s:%d: no final line or newline", name, lineno);
+                goto fail;
             }
-            if (fgets(buffer + length, bufsize - length, file) == NULL)
-                goto done;
             length = strlen(buffer);
+            p = buffer + length - 2;
         }
         if (length > 0)
             buffer[length - 1] = '\0';
@@ -399,65 +425,11 @@ read_conf_file(struct config *config, const char *name)
            handle include. */
         line = vector_split_space(buffer, NULL);
         if (line->count < 4) {
-            const char *included = line->strings[1];
-            struct stat st;
-
-            if (line->count != 2 || strcmp(line->strings[0], "include") != 0) {
-                syslog(LOG_ERR, "%s:%d: config parse error", name, lineno);
-                vector_free(line);
-                free(buffer);
-                fclose(file);
-                return -1;
-            }
-            if (strcmp(included, name) == 0) {
-                syslog(LOG_ERR, "%s:%d: %s recursively included", name,
-                       lineno, name);
-                vector_free(line);
-                free(buffer);
-                fclose(file);
-                return -1;
-            }
-            if (stat(included, &st) < 0) {
-                syslog(LOG_ERR, "%s:%d: included file %s not found", name,
-                       lineno, included);
-                vector_free(line);
-                free(buffer);
-                fclose(file);
-                return -1;
-            }
-            if (S_ISDIR(st.st_mode)) {
-                DIR *dir;
-                struct dirent *entry;
-
-                dir = opendir(included);
-                while ((entry = readdir(dir)) != NULL) {
-                    char *path;
-
-                    if (strchr(entry->d_name, '.') != NULL)
-                        continue;
-                    length = strlen(included) + 1 + strlen(entry->d_name) + 1;
-                    path = smalloc(length);
-                    snprintf(path, length, "%s/%s", included, entry->d_name);
-                    if (read_conf_file(config, path) < 0) {
-                        closedir(dir);
-                        free(path);
-                        free(buffer);
-                        fclose(file);
-                        return -1;
-                    }
-                    free(path);
-                }
-                closedir(dir);
-                vector_free(line);
-            } else {
-                if (read_conf_file(config, included) < 0) {
-                    vector_free(line);
-                    fclose(file);
-                    free(buffer);
-                    return -1;
-                }
-                vector_free(line);
-            }
+            s = handle_include(line, name, lineno, read_conf_file, config);
+            if (s < 0)
+                goto fail;
+            vector_free(line);
+            line = NULL;
             continue;
         }
 
@@ -470,12 +442,9 @@ read_conf_file(struct config *config, const char *name)
             else
                 config->allocated *= 2;
             size = config->allocated * sizeof(struct confline *);
-            config->rules = srealloc(config->rules, size);
+            config->rules = xrealloc(config->rules, size);
         }
-        confline = smalloc(sizeof(struct confline));
-        memset(confline, 0, sizeof(struct confline));
-        config->rules[config->count] = confline;
-        config->count++;
+        confline = xcalloc(1, sizeof(struct confline));
         confline->line    = line;
         confline->type    = line->strings[0];
         confline->service = line->strings[1];
@@ -493,71 +462,138 @@ read_conf_file(struct config *config, const char *name)
         /* One more syntax error possibility here: a line that only has a
            logmask setting but no ACL files. */
         if (line->count <= arg_i) {
-            syslog(LOG_ERR, "%s:%d: config parse error", name, lineno);
-            fclose(file);
-            free(buffer);
-            return -1;
+            warn("%s:%d: config parse error", name, lineno);
+            goto fail;
         }
 
         /* Grab the list of ACL files. */
         count = line->count - arg_i + 1;
-        confline->acls = smalloc(count * sizeof(char *));
+        confline->acls = xmalloc(count * sizeof(char *));
         for (i = 0; i < line->count - arg_i; i++)
             confline->acls[i] = line->strings[i + arg_i];
         confline->acls[i] = NULL;
+
+        /* Success.  Put the configuration line in place. */
+        config->rules[config->count] = confline;
+        config->count++;
+        confline = NULL;
+        line = NULL;
     }
 
-    /* Free allocated memory and return success if fgets succeeded. */
-done:
+    /* Free allocated memory and return success. */
     free(buffer);
-    if (ferror(file)) {
-        fclose(file);
-        return -1;
-    }
     fclose(file);
     return 0;
+
+    /* Abort with an error. */
+fail:
+    if (dir != NULL)
+        closedir(dir);
+    if (line != NULL)
+        vector_free(line);
+    if (confline != NULL) {
+        if (confline->logmask != NULL)
+            cvector_free(confline->logmask);
+        free(confline);
+    }
+    free(buffer);
+    fclose(file);
+    return -2;
+}
+
+
+/*
+**  Check to see if a given principal is in a given file.  This function is
+**  recursive to handle included ACL files and only does a simple check to
+**  prevent infinite recursion, so be careful.  The first argument is the user
+**  to check, which is passed in as a void * so that acl_check_file and
+**  read_conf_file can share common include-handling code.
+**
+**  Returns 0 if the user is authorized, -1 if they aren't, and -2 on some
+**  sort of failure (such as failure to read a file or a syntax error).
+*/
+static int
+acl_check_file(void *data, const char *aclfile)
+{
+    const char *user = data;
+    FILE *file = NULL;
+    char buffer[BUFSIZ];
+    char *p;
+    int lineno, s;
+    size_t length;
+    struct vector *line = NULL;
+
+    file = fopen(aclfile, "r");
+    if (file == NULL) {
+        syswarn("cannot open ACL file %s", aclfile);
+        return -2;
+    }
+    lineno = 0;
+    while (fgets(buffer, sizeof(buffer), file) != NULL) {
+        lineno++;
+        length = strlen(buffer);
+        if (length >= sizeof(buffer) - 1) {
+            warn("%s:%d: ACL file line too long", aclfile, lineno);
+            goto fail;
+        }
+
+        /* Skip blank lines or commented-out lines and remove trailing
+         * whitespace. */
+        p = buffer + length - 1;
+        while (isspace((int) *p))
+            p--;
+        p[1] = '\0';
+        p = buffer;
+        while (isspace((int) *p))
+            p++;
+        if (*p == '\0' || *p == '#')
+            continue;
+
+        /* Parse the line. */
+        if (strchr(p, ' ') != NULL) {
+            line = vector_split_space(buffer, NULL);
+            s = handle_include(line, aclfile, lineno, acl_check_file, data);
+            vector_free(line);
+            line = NULL;
+            if (s != -1) {
+                fclose(file);
+                return s;
+            }
+        } else if (strcmp(p, user) == 0) {
+            fclose(file);
+            return 0;
+        }
+    }
+    return -1;
+
+fail:
+    if (line != NULL)
+        vector_free(line);
+    if (file != NULL)
+        fclose(file);
+    return -2;
 }
 
 
 /*
 **  Check to see if a given principal is in a given list of ACL files.  The
-**  list of ACL files should be NULL-terminated.  Empty lines and lines
-**  beginning with # in the ACL file are ignored.
+**  list of ACL files should be NULL-terminated.  Special-case "ANYUSER" as
+**  the first filename.
 **
 **  Returns 0 on success and -1 on failure.
 */
 static int
 acl_check(char *userprincipal, char **acls)
 {
-    char *buf;
-    char *line;
-    int authorized = 0;
-    int i = 0;
+    int status, i;
 
-    if (strcmp(acls[i], "ANYUSER") == 0)
+    if (strcmp(acls[0], "ANYUSER") == 0)
         return 0;
-
-    while (acls[i] != NULL) {
-        if ((buf = read_file(acls[i])) == NULL)
-            return -1;
-
-        for (line = strtok(buf, "\n"); line != NULL; line = strtok(NULL, "\n")){
-            if (*line == '\0' || *line == '#')
-                continue;
-            if (strcmp(line, userprincipal) == 0) {
-                authorized = 1;
-                break;
-            }
-        }
-        
-        free(buf);
-
-        if (authorized)
-            return 0;
-        else
-            i++;
+    for (i = 0; acls[i] != NULL; i++) {
+        status = acl_check_file(userprincipal, acls[i]);
+        if (status != -1)
+            return status;
     }
-
     return -1;
 }
 
@@ -592,11 +628,10 @@ process_request(gss_ctx_id_t context, char *ret_message)
     req_argc = ntohl(network_order);
     cp += sizeof(OM_uint32);
 
-    if (verbose)
-        syslog(LOG_DEBUG, "argc is: %d\n", req_argc);
+    debug("argc is %d", req_argc);
     if (req_argc <= 0 || req_argc > MAXCMDARGS) {
         strcpy(ret_message, "Invalid argc in request message\n");
-        syslog(LOG_DEBUG, ret_message);
+        warn("invalid argc in request message");
         return NULL;
     }
         
@@ -612,9 +647,9 @@ process_request(gss_ctx_id_t context, char *ret_message)
         if (arglength > MAXBUFFER
             || arglength <= 0
             || cp + arglength > msg + msglength) {
-            strcpy (ret_message, 
-                    "Data unpacking error in getting arguments\n");
-            syslog(LOG_DEBUG, ret_message);
+            strcpy(ret_message, 
+                   "Data unpacking error in getting arguments\n");
+            warn("data unpacking error in getting arguments");
             return NULL;
         }
         memcpy(argbuf, cp, arglength);
@@ -648,7 +683,7 @@ process_response(gss_ctx_id_t context, OM_uint32 code, char *blob)
 
     flags = TOKEN_DATA;
 
-    msg = smalloc((2 * sizeof(OM_uint32) + bloblength));
+    msg = xmalloc((2 * sizeof(OM_uint32) + bloblength));
     msglength = bloblength + 2 * sizeof(OM_uint32);
 
     network_order = htonl(code);
@@ -674,7 +709,6 @@ static void
 log_command(struct vector *argvector, struct confline *cline,
             char *userprincipal)
 {
-    char log_message[MAXBUFFER];
     char *command;
     unsigned int i, j;
 
@@ -695,12 +729,7 @@ log_command(struct vector *argvector, struct confline *cline,
     } else {
         command = vector_join(argvector, " ");
     }
-
-    /* This block logs the requested command. */
-    snprintf(log_message, MAXBUFFER, "COMMAND from %s: ", userprincipal);
-    strncat(log_message, command, MAXBUFFER - strlen(log_message));
-    strcat(log_message, "\n");
-    syslog(LOG_INFO, "%s", log_message);
+    notice("COMMAND from %s: %s", userprincipal, command);
     free(command);
 }
 
@@ -738,8 +767,8 @@ process_command(struct config *config, struct vector *argvector,
     char scprincipal[100];
     int ret_length;
     char err_message[MAXBUFFER];
-    OM_uint32 ret_code;
-    int pid, pipebuffer;
+    int ret_code, pipebuffer;
+    pid_t pid;
     unsigned int i;
 
     memset(ret_message, 0, MAXBUFFER);
@@ -781,11 +810,12 @@ process_command(struct config *config, struct vector *argvector,
         snprintf(ret_message, MAXBUFFER,
                  "Access denied: %s principal not on the acl list\n",
                  userprincipal);
+        notice("access denied: %s not on ACL list", userprincipal);
         goto done;
     }
 
     /* Assemble the argv, envp, and fork off the child to run the command. */
-    req_argv = smalloc((argvector->count + 1) * sizeof(char *));
+    req_argv = xmalloc((argvector->count + 1) * sizeof(char *));
 
     /* Get the real program name, and use it as the first argument in argv
        passed to the command. */
@@ -804,7 +834,7 @@ process_command(struct config *config, struct vector *argvector,
        actually runs the command. */
     if (pipe(stdout_pipe) != 0 || pipe(stderr_pipe) != 0) {
         strcpy(ret_message, "Can't create pipes\n");
-        syslog(LOG_ERR, "%s", ret_message);
+        syswarn("cannot create pipes");
         ret_code = -1;
         goto done;
     }
@@ -812,7 +842,7 @@ process_command(struct config *config, struct vector *argvector,
     switch (pid = fork()) {
     case -1:
         strcpy(ret_message, "Can't fork\n");
-        syslog(LOG_ERR, "%s", ret_message);
+        syswarn("cannot fork");
         ret_code = -1;
         goto done;
 
@@ -833,7 +863,7 @@ process_command(struct config *config, struct vector *argvector,
         snprintf(remuser, MAXBUFFER, "REMUSER=%s", userprincipal);
         if (putenv(remuser) < 0) {
             strcpy(ret_message, "Can't set REMUSER environment variable\n");
-            syslog(LOG_ERR, "%s%m", ret_message);
+            syswarn("cannot set REMUSER environment variable");
             fprintf(stderr, ret_message);
             exit(-1);
         }
@@ -843,7 +873,7 @@ process_command(struct config *config, struct vector *argvector,
         if (putenv(scprincipal) < 0) {
             strcpy(ret_message,
                    "Can't set SCPRINCIPAL environment variable\n");
-            syslog(LOG_ERR, "%s%m", ret_message);
+            syswarn("cannot set SCPRINCIPAL environment variable");
             fprintf(stderr, ret_message);
             exit(-1);   
         }
@@ -854,8 +884,7 @@ process_command(struct config *config, struct vector *argvector,
            the errno information back up the stderr pipe to the parent, and
            dies. */
         strcpy(ret_message, strerror(errno));
-        fprintf(stderr, "%s\n", ret_message);
-        exit(-1);
+        sysdie("exec failed");
 
     /* In the parent. */
     default:
@@ -883,9 +912,8 @@ process_command(struct config *config, struct vector *argvector,
         waitpid(pid, &ret_code, 0);
 
         /* This does the crazy >>8 thing to get the real error code. */
-        if (WIFEXITED(ret_code)) {
+        if (WIFEXITED(ret_code))
             ret_code = (signed int) WEXITSTATUS(ret_code);
-        }
     }
 
  done:
@@ -915,7 +943,7 @@ process_connection(struct config *config, gss_cred_id_t server_creds)
     gss_buffer_desc client_name;
     gss_ctx_id_t context;
     OM_uint32 maj_stat, min_stat;
-    int ret_flags;
+    OM_uint32 ret_flags;
     struct vector *req_argv;
     OM_uint32 ret_code;
     char ret_message[MAXBUFFER];
@@ -934,18 +962,17 @@ process_connection(struct config *config, gss_cred_id_t server_creds)
         return -1;
 
     if (context == GSS_C_NO_CONTEXT) {
-        syslog(LOG_ERR, "Unauthenticated connection.\n");
+        warn("unauthenticated connection");
         return -1;
     }
 
     /* This is who just connected to us. */
-    userprincipal = smalloc(client_name.length + 1);
+    userprincipal = xmalloc(client_name.length + 1);
     memcpy(userprincipal, client_name.value, client_name.length);
     userprincipal[client_name.length] = '\0';
 
-    if (verbose)
-        syslog(LOG_DEBUG, "Accepted connection from \"%.*s\"\n",
-               (int)client_name.length, (char *)client_name.value);
+    debug("accepted connection from %.*s",
+          (int) client_name.length, (char *) client_name.value);
     
     gss_release_buffer(&min_stat, &client_name);
 
@@ -984,9 +1011,10 @@ process_connection(struct config *config, gss_cred_id_t server_creds)
 int
 main(int argc, char *argv[])
 {
-    char service_name[500];
+    char *service_name = NULL;
     char host_name[500];
-    struct hostent* hostinfo;
+    int option;
+    struct hostent *hostinfo;
     gss_cred_id_t server_creds;
     OM_uint32 min_stat;
     unsigned short port = 4444;
@@ -995,74 +1023,66 @@ main(int argc, char *argv[])
     const char *conffile = CONFIG_FILE;
     struct config *config;
 
-    service_name[0] = '\0';
-    verbose = 0;
-    use_syslog = 1;
-
-    /* Since we are called from tcpserver, prevent clients from holding on to 
-       us forever, and die after an hour. */
+    /* Since we are normally called from tcpserver or inetd, prevent clients
+       from holding on to us forever by dying after an hour. */
     alarm(60 * 60);
+
+    /* Establish identity and set up logging. */
+    message_program_name = "remctld";
     openlog("remctld", LOG_PID | LOG_NDELAY, LOG_DAEMON);
+    message_handlers_notice(1, message_log_syslog_info);
+    message_handlers_warn(1, message_log_syslog_warning);
+    message_handlers_die(1, message_log_syslog_err);
 
-    argc--;
-    argv++;
-    while (argc) {
-        if (strcmp(*argv, "-p") == 0) {
-            argc--;
-            argv++;
-            if (!argc)
-                usage();
-            port = atoi(*argv);
-        } else if (strcmp(*argv, "-v") == 0) {
-            verbose = 1;
-        } else if (strcmp(*argv, "-f")    == 0) {
-            argc--;
-            argv++;
-            if (!argc)
-                usage();
-            conffile = *argv;
-        } else if (strcmp(*argv, "-s")    == 0) {
-            argc--;
-            argv++;
-            if (!argc)
-                usage();
-            strncpy(service_name, *argv, sizeof(service_name));
-        } else if (strcmp(*argv, "-m")   == 0) {
+    /* Parse options. */
+    while ((option = getopt(argc, argv, "df:hmp:s:v")) != EOF) {
+        switch (option) {
+        case 'd':
+            message_handlers_debug(1, message_log_syslog_debug);
+            break;
+        case 'f':
+            conffile = optarg;
+            break;
+        case 'h':
+            usage(0);
+            break;
+        case 'm':
             do_standalone = 1;
-        } else
+            break;
+        case 'p':
+            port = atoi(optarg);
+            break;
+        case 's':
+            service_name = optarg;
+            break;
+        case 'v':
+            printf("remctld %s\n", PACKAGE_VERSION);
+            exit(0);
+            break;
+        default:
+            usage(1);
             break;
-        argc--;
-        argv++;
-    }
-
-    config = smalloc(sizeof(struct config));
-    if (read_conf_file(config, conffile) != 0) {
-        syslog(LOG_ERR, "%s%m\n", "Can't read conf file: ");
-        exit(1);
-    }
-
-    if (strlen(service_name) == 0) {
-        if (gethostname(host_name, sizeof(host_name)) < 0) {
-            syslog(LOG_ERR, "%s%m\n", "Can't get hostname: ");
-            exit(1);
         }
-        
-        if ((hostinfo = gethostbyname(host_name)) == NULL) {
-            syslog(LOG_ERR, "%s%s %m\n", "Can't resolve given hostname: ", 
-                   host_name);
-            exit(1);
-        }
-
-        /* service_name is for example host/zver.stanford.edu */
-        lowercase(hostinfo->h_name);
-        strcpy(service_name, "host/");
-        strcat(service_name, hostinfo->h_name);
     }
 
-    if (verbose)
-        syslog(LOG_DEBUG, "service_name: %s\n", service_name);
+    /* Read the configuration file. */
+    config = xcalloc(1, sizeof(struct config));
+    if (read_conf_file(config, conffile) != 0)
+        die("cannot read configuration file %s", conffile);
 
     /* This specifies the service name, checks out the keytab, etc. */
+    if (service_name == NULL) {
+        if (gethostname(host_name, sizeof(host_name)) < 0)
+            sysdie("cannot get hostname");
+        hostinfo = gethostbyname(host_name);
+        if (hostinfo == NULL)
+            die("cannot resolve hostname %s", host_name);
+        service_name = xmalloc(strlen("host/") + strlen(hostinfo->h_name) + 1);
+        strcpy(service_name, "host/");
+        strcat(service_name, hostinfo->h_name);
+        lowercase(service_name);
+    }
+    debug("service name: %s", service_name);
     if (server_acquire_creds(service_name, &server_creds) < 0)
         return -1;
 
@@ -1072,18 +1092,18 @@ main(int argc, char *argv[])
         if ((stmp = create_socket(port)) >= 0) {
             do {
                 /* Accept a TCP connection. */
-                if ((s = accept(stmp, NULL, 0)) < 0) {
-                    perror("accepting connection");
+                s = accept(stmp, NULL, 0);
+                if (s < 0) {
+                    syswarn("error accepting connection");
                     continue;
                 }
-
                 READFD = s;
                 WRITEFD = s;
-
                 if (process_connection(config, server_creds) < 0)
                     close(s);
             } while (1);
         }
+
     /* Here we read from stdin, and write to stdout to communicate with the
        network. */
     } else {
@@ -1094,8 +1114,8 @@ main(int argc, char *argv[])
         close(1);
     }
 
+    /* Clean up and exit.  We only reach here in regular mode. */
     gss_release_cred(&min_stat, &server_creds);
-
     return 0;
 }
 
